/*
 * Copyright 1993-2014 NVIDIA Corporation.  All rights reserved.
 *
 * Please refer to the NVIDIA end user license agreement (EULA) associated
 * with this source code for terms and conditions that govern your use of
 * this software. Any use, reproduction, disclosure, or distribution of
 * this software and related documentation outside the terms of the EULA
 * is strictly prohibited.
 *
 */

// This sample is an implementation of a simple line-of-sight algorithm:
// Given a height map and a ray originating at some observation point,
// it computes all the points along the ray that are visible from the
// observation point.
// It is based on the description made in "Guy E. Blelloch.  Vector models
// for data-parallel computing. MIT Press, 1990"

#ifdef _WIN32
#define NOMINMAX
#endif

// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>

// includes, project
#include "helper_functions.h"
#include "helper_cuda.h"
#include "helper_math.h"



////////////////////////////////////////////////////////////////////////////////
// declaration, types

// Boolean
typedef unsigned char Bool;
enum
{
    False = 0,
    True = 1
};

// 2D height field
struct HeightField
{
    int     width;
    float  *height;
};

// Ray
struct Ray
{
    float3 origin;
    float2 dir;
    int    length;
    float  oneOverLength;
};

////////////////////////////////////////////////////////////////////////////////
// declaration, variables

// Height field texture reference
texture<float, 2, cudaReadModeElementType> g_HeightFieldTex;

////////////////////////////////////////////////////////////////////////////////
// declaration, forward
int runTest(int argc, char **argv);
__global__ void computeAngles_kernel(const Ray, float *);
__global__ void computeVisibilities_kernel(const float *, const float *, int, Bool *);
__global__ void inclusive_scan_Hillis_Steele(const Ray, float *, float *);
__global__ void inclusive_scan_up_sweep(const Ray, float *);
__global__ void inclusive_scan_down_sweep(const Ray, float *);
void lineOfSight_CPU(const HeightField, const Ray, Bool *);
__device__ __host__ float2 getLocation(const Ray, int);
__device__ __host__ float getAngle(const Ray, float2, float);
void convert_heightField_bitmap(HeightField, const Ray, Bool *, int, int, LPCTSTR);
void SaveBitmapToFile( BYTE*, LONG , LONG , WORD ,  LPCTSTR  );

////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
int
main(int argc, char **argv)
{
    int res = runTest(argc, argv);

    cudaDeviceReset();

    if (res != 1)
    {
        printf("Test failed!\n");
        exit(EXIT_FAILURE);
    }

    printf("Test passed\n");
    exit(EXIT_SUCCESS);

}

////////////////////////////////////////////////////////////////////////////////
//! Run a line-of-sight test for CUDA
////////////////////////////////////////////////////////////////////////////////
int runTest(int argc, char **argv)
{
    ////////////////////////////////////////////////////////////////////////////
    // Device initialization

    printf("[%s] - Starting...\n", argv[0]);

    // use command-line specified CUDA device, otherwise use device with highest Gflops/s
    findCudaDevice(argc, (const char **)argv);

    ////////////////////////////////////////////////////////////////////////////
    // Timer

    // Create
    StopWatchInterface *timer;
    sdkCreateTimer(&timer);

    // Number of iterations to get accurate timing
    uint numIterations = 100;

    ////////////////////////////////////////////////////////////////////////////
    // Height field
	HeightField heightField;

    // Allocate in host memory the heightfield
    int2 dim = make_int2(1024, 1024);
    heightField.width = dim.x;
	unsigned int size_height = dim.x * dim.y;
    unsigned int mem_size_height = sizeof(float) * size_height;
    float *height = (float *)malloc(mem_size_height);
    cudaMallocHost((void **) &height,mem_size_height);
    heightField.height = (float *)&height[0];

    
	
    // Fill in with an arbitrary sine surface
    for (int x = 0; x < dim.x; ++x)
        for (int y = 0; y < dim.y; ++y)
        {
			float amp = 0.1f * (x + y);
            float period = 400.0f;
            *(heightField.height + dim.x * y + x) =
                amp * (sinf(sqrtf((float)(x * x + y * y)) * 2.0f * 3.1416f / period) + 1.0f);
		}

		

    // Allocate CUDA array in device memory ( CUDA Texture)
    cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc(32, 0, 0, 0, cudaChannelFormatKindFloat);
    cudaArray *heightFieldArray;
    cudaMallocArray(&heightFieldArray, &channelDesc, dim.x, dim.y);

    // Initialize device memory
    cudaMemcpyToArray(heightFieldArray, 0, 0, heightField.height,dim.x * dim.y * sizeof(float), cudaMemcpyHostToDevice);

    // Set texture parameters
    g_HeightFieldTex.addressMode[0] = cudaAddressModeClamp;
    g_HeightFieldTex.addressMode[1] = cudaAddressModeClamp;
    g_HeightFieldTex.filterMode = cudaFilterModePoint;
    g_HeightFieldTex.normalized = 0;

    // Bind CUDA array to texture reference
    cudaBindTextureToArray(g_HeightFieldTex, heightFieldArray,channelDesc);

    ////////////////////////////////////////////////////////////////////////////
    // Ray (starts at origin and traverses the height field diagonally)
	Ray ray;
    ray.origin = make_float3(0, 0, 2.0f);
    int2 dir = make_int2(dim.x - 1, dim.y - 1);
    ray.dir = make_float2((float)dir.x, (float)dir.y);
	int devID;
    cudaGetDevice( &devID );
    cudaDeviceProp props;
    cudaGetDeviceProperties( &props, devID );
	//set the ray length according to capability of device
	ray.length = props.maxThreadsPerBlock;
    ray.oneOverLength = 1.0f / ray.length;


    ////////////////////////////////////////////////////////////////////////////
    // View angles

    // Allocate view angles for each point along the ray
	float *d_angles;
	size_t mem_size_angles = ray.length * sizeof(float);
    cudaMalloc((void **) &d_angles, mem_size_angles);

    // Allocate result of max-scan operation on the array of view angles ( d_scannedBuffer is used for double buffering in case inclusive scan using Hillis-Steele algorithm )
    float *d_scannedAngles, *d_scannedAnglesBuffer;
	size_t mem_size_scannedAngles = ray.length * sizeof(float);
    cudaMalloc((void **) &d_scannedAngles, mem_size_scannedAngles);
	cudaMalloc((void **) &d_scannedAnglesBuffer, mem_size_scannedAngles);

    ////////////////////////////////////////////////////////////////////////////
    // Visibility results

    // Allocate visibility results for each point along the ray
	Bool *h_visibilitiesRef, *d_visibilities, *h_visibilities;
	size_t mem_size_visibilities = ray.length * sizeof(Bool);

	//device visibilites array
    cudaMalloc((void **) &d_visibilities, mem_size_visibilities);
	//host visibilites array
	cudaMallocHost((void **) &h_visibilities, mem_size_visibilities);
	//host visibilites array used for storing the reference visibilities
	cudaMallocHost((void **) &h_visibilitiesRef, mem_size_visibilities);

    ////////////////////////////////////////////////////////////////////////////
    // Device solution

    // Execution configuration
    dim3 block(ray.length);
    dim3 grid(1);

    // Compute device solution
    printf("Line of sight\n");
    sdkStartTimer(&timer);

    for (uint i = 0; i < numIterations; ++i)
    {

        // Compute view angle for each point along the ray
        computeAngles_kernel<<<grid, block>>>(ray, d_angles);

		//Copy the view anlges to d_scannedAngles device memory
		cudaMemcpy(d_scannedAngles, d_angles, mem_size_scannedAngles, cudaMemcpyDeviceToDevice);
		cudaMemcpy(d_scannedAnglesBuffer, d_angles, mem_size_scannedAngles, cudaMemcpyDeviceToDevice);


        // Perform a max-scan operation on the array of view angles
        inclusive_scan_Hillis_Steele<<<grid, block>>>(ray, d_scannedAngles, d_scannedAnglesBuffer);

		//b> TODO or instead use Blellochs inclusive_scan_up_sweep and inclusive_scan_down_sweep kernels to perform max scan operation
    

        // Compute visibility results based on the array of view angles
        // and its scanned version
        computeVisibilities_kernel<<<grid, block>>>(d_angles, d_scannedAngles, ray.length, d_visibilities);


    }

    cudaDeviceSynchronize();
    sdkStopTimer(&timer);
 

    // Copy visibility results back to the host
    cudaMemcpy(h_visibilities, d_visibilities, mem_size_visibilities, cudaMemcpyDeviceToHost);

	 ////////////////////////////////////////////////////////////////////////////
    // Reference solution
    lineOfSight_CPU(heightField, ray, h_visibilitiesRef);

	// generates gray scale bitmap image along with visible points along ray colored as blue and non-visible points as red
	convert_heightField_bitmap(heightField, ray, h_visibilities, dim.x, dim.y, "height_field.bmp");


    // Compare device visibility results against reference results
    bool res = compareData(h_visibilitiesRef, h_visibilities, ray.length, 0.0f, 0.0f);
    printf("Average time: %f ms\n\n", sdkGetTimerValue(&timer) / numIterations);
    sdkResetTimer(&timer);

    // Cleanup memory
    cudaFreeArray(heightFieldArray);
    return res;
}

////////////////////////////////////////////////////////////////////////////////
//! Compute view angles for each point along the ray
//! @param ray         ray
//! @param angles      view angles
////////////////////////////////////////////////////////////////////////////////
__global__ void computeAngles_kernel(const Ray ray, float *angles)
{
    uint i = blockDim.x * blockIdx.x + threadIdx.x;

    if (i < ray.length)
    {
        float2 location = getLocation(ray, i + 1);
        float height = tex2D(g_HeightFieldTex, location.x, location.y);
        float angle = getAngle(ray, location, height);
        angles[i] = angle;
    }
}

////////////////////////////////////////////////////////////////////////////////
//! Compute inclusive scan vector for the view angles vector Hilles Steele Algorithm
//! @param ray                      ray
//! @param scannedAngles            max scan operation applied angles ( where view angles are stored when the kernel is called) 
//! @param scannedAnglesBuffer      max scan operation applied angles buffer used for double buferring in Hillis Steele Algorithm 
////////////////////////////////////////////////////////////////////////////////

/* a> TODO by the student impelment Hillis Steele Algorithm */
__global__ void inclusive_scan_Hillis_Steele(const Ray ray, float *scannedAngles, float *scannedAnglesBuffer)
{

}


///////////////////////////////////////////////////////////////////////////////
//!  Blelloch Algorithm up sweep step
//! @param ray                      ray
//! @param scannedAngles            max scan operation applied angles  (where view angles are stored when the kernel is called)
////////////////////////////////////////////////////////////////////////////////

/* b> TODO by the student implement Blelloch Algorithm up sweep step */
__global__ void inclusive_scan_up_sweep(const Ray ray, float *scannedAngles)
{
  

}


///////////////////////////////////////////////////////////////////////////////
//! Blelloch Algorithm down sweep step
//! @param ray                      ray
//! @param scannedAngles            max scan operation applied angles  (where scanned angles from previous up sweep step are stored when the kernel is called)
////////////////////////////////////////////////////////////////////////////////

/* b> TODO by the student implement Blelloch Algorithm down sweep step */
__global__ void inclusive_scan_down_sweep(const Ray ray, float *scannedAngles)
{
  

}


////////////////////////////////////////////////////////////////////////////////
//! Compute visibility for each point along the ray
//! @param angles          view angles
//! @param scannedAngles   max-scanned view angles
//! @param numAngles       number of view angles
//! @param visibilities    boolean array indicating the visibility of each point
//!                        along the ray
////////////////////////////////////////////////////////////////////////////////
__global__ void computeVisibilities_kernel(const float *angles,
                                           const float *scannedAngles,
                                           int numAngles,
                                           Bool *visibilities)
{
    uint i = blockDim.x * blockIdx.x + threadIdx.x;

    if (i < numAngles)
    {
        visibilities[i] = scannedAngles[i] <= angles[i];
    }
}

////////////////////////////////////////////////////////////////////////////////
//! Compute reference data set
//! @param heightField     height field
//! @param ray             ray
//! @param visibilities    boolean array indicating the visibility of each point
//!                        along the ray
////////////////////////////////////////////////////////////////////////////////
void lineOfSight_CPU(const HeightField heightField, const Ray ray,
                      Bool *visibilities)
{
    float angleMax = asinf(-1.0f);

    for (int i = 0; i < ray.length; ++i)
    {
        float2 location = getLocation(ray, i + 1);
        float height = *(heightField.height
                         + heightField.width * (int)floorf(location.y)
                         + (int)floorf(location.x));
        float angle = getAngle(ray, location, height);

        if (angle > angleMax)
        {
            angleMax = angle;
            visibilities[i] = True;
        }
        else
        {
            visibilities[i] = False;
        }
		
    }
}

////////////////////////////////////////////////////////////////////////////////
//! Compute the 2D coordinates of the point located at i steps from the origin
//! of the ray
//! @param ray      ray
//! @param i        integer offset along the ray
////////////////////////////////////////////////////////////////////////////////
__device__ __host__ float2 getLocation(const Ray ray, int i)
{
    float step = i * ray.oneOverLength;
    return make_float2(ray.origin.x, ray.origin.y) + ray.dir * step;
}

////////////////////////////////////////////////////////////////////////////////
//! Compute the angle of view between a 3D point and the origin of the ray
//! @param ray        ray
//! @param location   2D coordinates of the input point
//! @param height     height of the input point
////////////////////////////////////////////////////////////////////////////////
__device__ __host__ float getAngle(const Ray ray, float2 location, float height)
{
    float2 dir = location - make_float2(ray.origin.x, ray.origin.y);
    return atanf((height - ray.origin.z) / length(dir));
}



////////////////////////////////////////////////////////////////////////////////
// Other Utility functions
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
//! convert the height field to bitmap file
//! @param heightField     height field
//! @param ray             ray
//! @param visibilities    boolean array indicating the visibility of each point
//!                        along the ray
//! @param image_width     width of image to be generated
//! @param image_height    height of image to be generated
//! @param fileName        name of bitmap file to be generated
////////////////////////////////////////////////////////////////////////////////

void convert_heightField_bitmap(HeightField heightField, Ray ray, Bool *visibilities, int image_width, int image_height, LPCTSTR fileName)
{

    BYTE* buf = new BYTE[ 3 * image_width * image_height ];  
      
    int c = 0;  
     

	//computes the contours of the height field for values 50, 100, 150, 200
    for ( int i = 0; i < image_height; i++ )  
    {  
        for ( int j = 0; j < image_width; j++ )  
        {  
			int val;
			val = (int)(((heightField.height[ i + j*image_width ]) * 255)/(409.6f));
			 buf[ c + 0 ] = (BYTE) val;  
            buf[ c + 1 ] = (BYTE) val;  
            buf[ c + 2 ] = (BYTE) val;  
      
            c += 3;  
        }  
    }

	//write the pixels with blue for visible points and red for non-visible points along the ray
	for ( int i = 0; i < ray.length; i++)
	{
		float2 currentPoint = getLocation(ray,i);
		int y = (int)floorf(currentPoint.y);
		int x = (int)floorf(currentPoint.x);
		int r = 0;
		int g = 0;
		int b = 0;
		if(visibilities[i])
		{
			r=0;g=255;b=255;
			buf[ y * image_width * 3 + x * 3 + 0 ] = (BYTE) b;  
			buf[ y * image_width * 3 + x * 3 + 1 ] = (BYTE) g;  
			buf[ y * image_width * 3 + x * 3 + 2 ] = (BYTE) r;
		}
		else
		{
			r=255;g=0;b=0;
			buf[ y * image_width * 3 + x * 3 + 0 ] = (BYTE) b;  
			buf[ y * image_width * 3 + x * 3 + 1 ] = (BYTE) g;  
			buf[ y * image_width * 3 + x * 3 + 2 ] = (BYTE) r;
		}
	}


      
    SaveBitmapToFile( (BYTE*) buf,  
                                image_width,  
                                image_height,  
                                24,  
                                fileName );  
      
    delete [] buf;  
}


////////////////////////////////////////////////////////////////////////////////
//! saves the heightfield as a bitmap( consisting of contours)
////////////////////////////////////////////////////////////////////////////////
void SaveBitmapToFile( BYTE* pBitmapBits,  
                                       LONG lWidth,  
                                       LONG lHeight,  
                                       WORD wBitsPerPixel,  
                                       LPCTSTR lpszFileName )  
{  
    // Some basic bitmap parameters  
    unsigned long headers_size = sizeof( BITMAPFILEHEADER ) +  
                                 sizeof( BITMAPINFOHEADER );  
    unsigned long padding_size    = ( 4 - ( ( lWidth * 3 ) % 4 ) ) % 4;  
    unsigned long pixel_data_size = lHeight * ( ( lWidth * 3 ) + padding_size );  
  
    BITMAPINFOHEADER bmpInfoHeader = {0};  
  
    // Set the size  
    bmpInfoHeader.biSize = sizeof(BITMAPINFOHEADER);  
  
    // Bit count  
    bmpInfoHeader.biBitCount = wBitsPerPixel;  
  
    // Use all colors  
    bmpInfoHeader.biClrImportant = 0;  
  
    // Use as many colors according to bits per pixel  
    bmpInfoHeader.biClrUsed = 0;  
  
    // Store as un Compressed  
    bmpInfoHeader.biCompression = BI_RGB;  
  
    // Set the height in pixels  
    bmpInfoHeader.biHeight = lHeight;  
  
    // Width of the Image in pixels  
    bmpInfoHeader.biWidth = lWidth;  
  
    // Default number of planes  
    bmpInfoHeader.biPlanes = 1;  
  
    // Calculate the image size in bytes  
    bmpInfoHeader.biSizeImage = pixel_data_size;  
  
    BITMAPFILEHEADER bfh = {0};  
  
    // This value should be values of BM letters i.e 0x4D42  
    // 0x4D = M 042 = B storing in reverse order to match with endian  
    bfh.bfType=0x4D42;  
    /* or  bfh.bfType = B+(M << 8); 
    // <<8 used to shift M to end  */  
  
    // Offset to the RGBQUAD  
    bfh.bfOffBits = headers_size;  
  
    // Total size of image including size of headers  
    bfh.bfSize =  headers_size + pixel_data_size;  
  
    // Create the file in disk to write  
    HANDLE hFile = CreateFile( lpszFileName,  
                               GENERIC_WRITE,  
                               0,  
                               NULL,  
                               CREATE_ALWAYS,  
                               FILE_ATTRIBUTE_NORMAL,  
                               NULL );  
  
    // Return if error opening file  
    if( !hFile ) return;  
  
    DWORD dwWritten = 0;  
  
    // Write the File header  
    WriteFile( hFile,  
               &bfh,  
               sizeof(bfh),  
               &dwWritten ,  
               NULL );  
  
    // Write the bitmap info header  
    WriteFile( hFile,  
               &bmpInfoHeader,  
               sizeof(bmpInfoHeader),  
               &dwWritten,  
               NULL );  
  
    // Write the RGB Data  
    WriteFile( hFile,  
               pBitmapBits,  
               bmpInfoHeader.biSizeImage,  
               &dwWritten,  
               NULL );  
  
    // Close the file handle  
    CloseHandle( hFile );  
}  


////////////////////////////////////////////////////////////////////////////////
//! General CUDA error handler
//! Prints error message, if something went wrong
//! @param line         line number were the method is called
////////////////////////////////////////////////////////////////////////////////

void handleCUDAerror( int line )
{
	cudaError_t err = cudaGetLastError();
	if ( err == cudaSuccess )
		return;

	printf("CUDA error message: %s\n"
		   "line %d\n", cudaGetErrorString(err), line );

	int devID;
    cudaGetDevice( &devID );
    cudaDeviceProp props;
    cudaGetDeviceProperties( &props, devID );
    printf("CUDA device %d: \"%s\" with compute capability %d.%d\n",
           devID, props.name, props.major, props.minor);
}


